# Generated by Django 2.1 on 2018-10-06 20:57

import edc_sites.models
from django.db import migrations

import edc_registration.models.managers


class Migration(migrations.Migration):

    dependencies = [("edc_registration", "0014_auto_20180814_2017")]

    operations = [
        migrations.AlterModelManagers(
            name="registeredsubject",
            managers=[
                ("on_site", edc_sites.models.CurrentSiteManager()),
                (
                    "objects",
                    edc_registration.models.managers.RegisteredSubjectManager(),
                ),
            ],
        )
    ]



import re

from django.core.exceptions import ObjectDoesNotExist
from django.core.validators import RegexValidator
from django.db import models, transaction
from django.utils.translation import gettext as _
from django_crypto_fields.fields import (
    EncryptedCharField,
    FirstnameField,
    IdentityField,
    LastnameField,
)
from edc_constants.choices import GENDER
from edc_constants.constants import UUID_PATTERN
from edc_identifier.model_mixins import UniqueSubjectIdentifierModelMixin
from edc_model import models as edc_models
from edc_model_fields.fields import IdentityTypeField, IsDateEstimatedField
from edc_sites.models import CurrentSiteManager, SiteModelMixin

from .managers import RegisteredSubjectManager


class RegisteredSubjectError(Exception):
    pass


class RegisteredSubject(
    UniqueSubjectIdentifierModelMixin, SiteModelMixin, edc_models.BaseUuidModel
):
    """A model mixin for the RegisteredSubject model (only)."""

    # may not be available when instance created (e.g. infants prior to birth
    # report)
    first_name = FirstnameField(null=True)

    # may not be available when instance created (e.g. infants or household
    # subject before consent)
    last_name = LastnameField(verbose_name="Last name", null=True)

    # may not be available when instance created (e.g. infants)
    initials = EncryptedCharField(
        validators=[
            RegexValidator(
                regex=r"^[A-Z]{2,3}$",
                message=(
                    "Ensure initials consist of letters " "only in upper case, no spaces."
                ),
            )
        ],
        null=True,
    )

    dob = models.DateField(
        verbose_name=_("Date of birth"),
        null=True,
        blank=False,
        help_text=_("Format is YYYY-MM-DD"),
    )

    is_dob_estimated = IsDateEstimatedField(
        verbose_name=_("Is date of birth estimated?"), null=True, blank=False
    )

    gender = models.CharField(
        verbose_name="Gender", max_length=1, choices=GENDER, null=True, blank=False
    )

    subject_consent_id = models.CharField(max_length=100, null=True, blank=True)

    registration_identifier = models.CharField(max_length=36, null=True, blank=True)

    sid = models.CharField(verbose_name="SID", max_length=15, null=True, blank=True)

    subject_type = models.CharField(max_length=25, null=True, blank=True)

    relative_identifier = models.CharField(
        verbose_name="Identifier of immediate relation",
        max_length=36,
        null=True,
        blank=True,
        help_text="For example, mother's identifier, if available / appropriate",
    )

    identity = IdentityField(null=True, blank=True)

    identity_type = IdentityTypeField(null=True, blank=True)

    screening_identifier = models.CharField(max_length=36, null=True, blank=True)

    screening_datetime = models.DateTimeField(null=True, blank=True)

    screening_age_in_years = models.IntegerField(null=True, blank=True)

    registration_datetime = models.DateTimeField(null=True, blank=True)

    # For simplicity, if going straight from screen to rando,
    # update both registration date and randomization date
    randomization_datetime = models.DateTimeField(null=True, blank=True)

    registration_status = models.CharField(
        verbose_name="Registration status", max_length=25, null=True, blank=True
    )

    consent_datetime = models.DateTimeField(null=True, blank=True)

    comment = models.TextField(verbose_name="Comment", max_length=250, null=True, blank=True)

    additional_key = models.CharField(
        max_length=36,
        verbose_name="-",
        editable=False,
        default=None,
        null=True,
        help_text=(
            "A uuid (or some other text value) to be added to bypass the "
            "unique constraint of just firstname, initials, and dob."
            "The default constraint proves limiting since the source "
            "model usually has some other attribute in additional to "
            "first_name, initials and dob which is not captured in "
            "this model"
        ),
    )

    dm_comment = models.CharField(
        verbose_name="Data Management comment",
        max_length=150,
        null=True,
        editable=False,
    )

    randomization_list_model = models.CharField(max_length=150, null=True)

    on_site = CurrentSiteManager()

    history = edc_models.HistoricalRecords()

    objects = RegisteredSubjectManager()

    def save(self, *args, **kwargs):
        if self.identity:
            self.additional_key = None
        self.set_uuid_as_subject_identifier_if_none()
        self.raise_on_duplicate("subject_identifier")
        self.raise_on_duplicate("identity")
        self.raise_on_changed_subject_identifier()
        super().save(*args, **kwargs)

    def natural_key(self):
        return (self.subject_identifier_as_pk,)  # noqa

    def __str__(self):
        return self.masked_subject_identifier

    natural_key.dependencies = ["sites.Site"]

    def update_subject_identifier_on_save(self):
        """Overridden to not set the subject identifier on save."""
        if not self.subject_identifier:
            self.subject_identifier = self.subject_identifier_as_pk.hex
        elif re.match(UUID_PATTERN, self.subject_identifier):
            pass
        return self.subject_identifier

    def make_new_identifier(self):
        return self.subject_identifier_as_pk.hex

    @property
    def masked_subject_identifier(self):
        """Returns the subject identifier, if set, otherwise
        the string '<identifier not set>'.
        """
        if not self.subject_identifier_is_set:
            return "<identifier not set>"
        return self.subject_identifier

    @property
    def subject_identifier_is_set(self):
        """Returns True if subject identifier has been set to a
        subject identifier; that is, no longer the default UUID.
        """
        is_set = True
        try:
            obj = self.__class__.objects.get(pk=self.id)
        except ObjectDoesNotExist:
            is_set = False
        else:
            if re.match(UUID_PATTERN, obj.subject_identifier):
                return False
        return is_set

    def raise_on_changed_subject_identifier(self):
        """Raises an exception if there is an attempt to change
        the subject identifier for an existing instance if the subject
        identifier is already set.
        """
        if self.id and self.subject_identifier_is_set:
            with transaction.atomic():
                obj = self.__class__.objects.get(pk=self.id)
                if obj.subject_identifier != self.subject_identifier_as_pk.hex:
                    if self.subject_identifier != obj.subject_identifier:
                        raise RegisteredSubjectError(
                            "Subject identifier cannot be changed for "
                            "existing registered subject. "
                            f"Got {self.subject_identifier} <> {obj.subject_identifier}."
                        )

    def raise_on_duplicate(self, attrname):
        """Checks if the subject identifier (or other attr) is in use,
        for new and existing instances.
        """
        if getattr(self, attrname):
            with transaction.atomic():
                error_msg = (
                    f"Cannot {{action}} registered subject with a duplicate "
                    f"'{attrname}'. Got {getattr(self, attrname)}."
                )
                try:
                    obj = self.__class__.objects.exclude(
                        **{"pk": self.pk} if self.id else {}
                    ).get(**{attrname: getattr(self, attrname)})
                    if not self.id:
                        raise RegisteredSubjectError(error_msg.format(action="insert"))
                    elif self.subject_identifier_is_set and obj.id != self.id:
                        raise RegisteredSubjectError(error_msg.format(action="update"))
                    else:
                        raise RegisteredSubjectError(error_msg.format(action="update"))
                except ObjectDoesNotExist:
                    pass

    def set_uuid_as_subject_identifier_if_none(self):
        """Inserts a random uuid as a dummy identifier for a new
        instance.

        Model uses subject_identifier_as_pk as a natural key for
        serialization/deserialization. Value must not change
        once set.
        """
        if not self.subject_identifier:
            self.subject_identifier = self.subject_identifier_as_pk.hex

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Registered Subject"
        verbose_name_plural = "Registered Subjects"
        ordering = ["subject_identifier"]
        unique_together = ("first_name", "dob", "initials", "additional_key")
        indexes = [
            models.Index(fields=["first_name", "dob", "initials", "additional_key"]),
            models.Index(fields=["identity", "subject_identifier", "screening_identifier"]),
        ]
        permissions = (
            ("display_firstname", "Can display first name"),
            ("display_lastname", "Can display last name"),
            ("display_dob", "Can display DOB"),
            ("display_identity", "Can display identity number"),
            ("display_initials", "Can display initials"),
        )



# Generated by Django 2.0 on 2017-12-28 18:16

import django.contrib.sites.managers
import django.db.models.deletion
import django.db.models.manager
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("sites", "0002_alter_domain_unique"),
        ("edc_registration", "0008_auto_20170810_1032"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="registeredsubject",
            managers=[
                ("objects", django.db.models.manager.Manager()),
                ("on_site", django.contrib.sites.managers.CurrentSiteManager()),
            ],
        ),
        migrations.RemoveField(model_name="registeredsubject", name="study_site"),
        migrations.AddField(
            model_name="registeredsubject",
            name="site",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="sites.Site",
            ),
        ),
    ]



# Generated by Django 3.2.4 on 2021-08-01 17:21

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_registration", "0021_auto_20200513_0023"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="historicalregisteredsubject",
            name="identity_or_pk",
        ),
        migrations.RemoveField(
            model_name="registeredsubject",
            name="identity_or_pk",
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.4 on 2017-02-09 07:24
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_registration", "0005_auto_20170111_1809")]

    operations = [
        migrations.AlterField(
            model_name="registeredsubject",
            name="additional_key",
            field=models.CharField(
                default=None,
                editable=False,
                help_text="A uuid (or some other text value) to be added to bypass the unique constraint of just firstname, initials, and dob.The default constraint proves limiting since the source model usually has some other attribute in additional to first_name, initials and dob which is not captured in this model",
                max_length=36,
                null=True,
                verbose_name="-",
            ),
        )
    ]



# Generated by Django 3.2.11 on 2022-04-12 18:51

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_registration", "0022_auto_20210801_2021"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="registeredsubject",
            options={
                "default_permissions": ("add", "change", "delete", "view", "export", "import"),
                "get_latest_by": "modified",
                "ordering": ["subject_identifier"],
                "permissions": (
                    ("display_firstname", "Can display first name"),
                    ("display_lastname", "Can display last name"),
                    ("display_dob", "Can display DOB"),
                    ("display_identity", "Can display identity number"),
                    ("display_initials", "Can display initials"),
                ),
                "verbose_name": "Registered Subject",
                "verbose_name_plural": "Registered Subjects",
            },
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.3 on 2016-12-05 08:04
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_registration", "0002_auto_20161127_2226")]

    operations = [
        migrations.AddField(
            model_name="registeredsubject",
            name="subject_type",
            field=models.CharField(blank=True, max_length=25, null=True),
        )
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.3 on 2016-11-27 22:26
from __future__ import unicode_literals

import edc_utils
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_registration", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="registeredsubject",
            name="created",
            field=models.DateTimeField(default=edc_utils.date.get_utcnow, editable=False),
        ),
        migrations.AlterField(
            model_name="registeredsubject",
            name="modified",
            field=models.DateTimeField(default=edc_utils.date.get_utcnow, editable=False),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.4 on 2016-12-21 00:18
from __future__ import unicode_literals

import edc_utils
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_registration", "0003_registeredsubject_subject_type")]

    operations = [
        migrations.AlterField(
            model_name="registeredsubject",
            name="subject_identifier",
            field=models.CharField(
                max_length=50, unique=True, verbose_name="Subject Identifier"
            ),
        ),
        migrations.AlterField(
            model_name="registeredsubject",
            name="subject_identifier_as_pk",
            field=models.CharField(
                default=edc_utils.get_uuid,
                editable=False,
                max_length=50,
                verbose_name="Subject Identifier as pk",
            ),
        ),
    ]



from django import forms
from django.core.exceptions import ObjectDoesNotExist
from edc_sites.forms import SiteModelFormMixin

from .utils import get_registered_subject_model_cls


class ModelFormSubjectIdentifierMixin(SiteModelFormMixin):

    subject_identifier = forms.CharField(
        label="Subject Identifier",
        required=False,
        widget=forms.TextInput(attrs={"readonly": "readonly"}),
        help_text="(read-only)",
    )

    def clean(self):
        cleaned_data = super().clean()
        subject_identifier = cleaned_data.get("subject_identifier")
        try:
            get_registered_subject_model_cls().objects.get(
                subject_identifier=subject_identifier
            )
        except ObjectDoesNotExist:
            raise forms.ValidationError({"subject_identifier": "Invalid."})
        return cleaned_data



from typing import Tuple

from django.contrib import admin
from django_audit_fields.admin import audit_fields
from edc_auth.auth_objects import PII, PII_VIEW
from edc_model_admin.dashboard import ModelAdminSubjectDashboardMixin


class RegisteredSubjectModelAdminMixin(ModelAdminSubjectDashboardMixin, admin.ModelAdmin):

    ordering = ("registration_datetime",)

    date_hierarchy = "registration_datetime"

    instructions = []

    @staticmethod
    def show_pii(request) -> bool:
        return request.user.groups.filter(name__in=[PII, PII_VIEW]).exists()

    def get_fieldsets(self, request, obj=None):
        """
        Hook for specifying fieldsets.
        """
        if self.fieldsets:
            if self.show_pii(request):
                return self.fieldsets
            else:
                return self.fieldsets_no_pii
        return [(None, {"fields": self.get_fields(request, obj)})]

    def get_readonly_fields(self, request, obj=None) -> Tuple[str, ...]:
        readonly_fields = super().get_readonly_fields(request, obj=obj)
        return (
            readonly_fields
            + (
                "subject_identifier",
                "sid",
                "first_name",
                "last_name",
                "initials",
                "dob",
                "gender",
                "subject_type",
                "registration_status",
                "identity",
                "screening_identifier",
                "screening_datetime",
                "registration_datetime",
                "randomization_datetime",
                "consent_datetime",
            )
            + audit_fields
        )

    def get_list_display(self, request):
        list_display = super().get_list_display(request)
        if self.show_pii(request):
            custom_fields = (
                "subject_identifier",
                "dashboard",
                "first_name",
                "initials",
                "gender",
                "subject_type",
                "sid",
                "registration_status",
                "site",
                "user_created",
                "created",
            )
        else:
            custom_fields = (
                "subject_identifier",
                "dashboard",
                "gender",
                "subject_type",
                "sid",
                "registration_status",
                "site",
                "user_created",
                "created",
            )
        return custom_fields + tuple(f for f in list_display if f not in custom_fields)

    def get_list_filter(self, request) -> Tuple[str, ...]:
        list_filter = super().get_list_filter(request)
        custom_fields = (
            "subject_type",
            "registration_status",
            "screening_datetime",
            "registration_datetime",
            "gender",
            "site",
            "hostname_created",
        )
        return custom_fields + tuple(f for f in list_filter if f not in custom_fields)

    def get_search_fields(self, request) -> Tuple[str, ...]:
        search_fields = super().get_search_fields(request)
        pii_fields = (
            "first_name",
            "initials",
            "identity",
        )
        search_fields += (
            "subject_identifier",
            "sid",
            "id",
            "screening_identifier",
            "registration_identifier",
        )
        if not self.show_pii(request):
            return tuple(set(f for f in search_fields if f not in pii_fields))
        return tuple(set(search_fields + pii_fields))



from edc_auth.auth_objects import PII, PII_VIEW
from edc_auth.site_auths import site_auths
from edc_export.auth_objects import EXPORT

site_auths.update_group("edc_registration.export_registeredsubject", name=EXPORT)
site_auths.update_group(
    "edc_registration.display_dob",
    "edc_registration.display_firstname",
    "edc_registration.display_identity",
    "edc_registration.display_initials",
    "edc_registration.display_lastname",
    "edc_registration.view_historicalregisteredsubject",
    "edc_registration.view_registeredsubject",
    name=PII,
)

site_auths.update_group(
    "edc_registration.display_dob",
    "edc_registration.display_firstname",
    "edc_registration.display_identity",
    "edc_registration.display_initials",
    "edc_registration.display_lastname",
    "edc_registration.view_historicalregisteredsubject",
    "edc_registration.view_registeredsubject",
    name=PII_VIEW,
)
site_auths.add_pii_model("edc_registration.registeredsubject")



from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from edc_model import DEFAULT_BASE_FIELDS


class UpdatesOrCreatesRegistrationModelError(Exception):
    pass


class UpdatesOrCreatesRegistrationModelMixin(models.Model):

    """A model mixin that creates or updates RegisteredSubject
    on post_save signal.
    """

    @property
    def registration_model(self):
        """Returns the RegisteredSubject model, Do not override."""
        return django_apps.get_app_config("edc_registration").model

    def registration_update_or_create(self):
        """Creates or Updates the registration model with attributes
        from this instance.

        Called from the signal.

        Note: `registration_unique_field` is typically
              "subject_identifier".
        """
        if not getattr(self, self.registration_unique_field):
            raise UpdatesOrCreatesRegistrationModelError(
                f"Cannot update or create RegisteredSubject. "
                f"Field value for '{self.registration_unique_field}' is None."
            )

        registration_value = getattr(self, self.registration_unique_field)
        registration_value = self.to_string(registration_value)
        try:
            obj = self.registration_model.objects.get(
                **{self.registered_subject_unique_field: registration_value}
            )
        except ObjectDoesNotExist:
            pass
        else:
            self.registration_raise_on_illegal_value_change(obj)
        registered_subject, created = self.registration_model.objects.update_or_create(
            **{self.registered_subject_unique_field: registration_value},
            defaults=self.registration_options,
        )
        return registered_subject, created

    def to_string(self, value):
        """Returns a string.

        Converts UUID to string using .hex.
        """
        try:
            value = str(value.hex)
        except AttributeError:
            pass
        return value

    @property
    def registration_unique_field(self):
        """Returns the field attr on YOUR model that will update
        `registered_subject_unique_field`.
        """
        return "subject_identifier"

    @property
    def registered_subject_unique_field(self):
        """Returns the field attr on THIS model, registered subject,
        to be queried against by the value of `registration_unique_field`.
        """
        return self.registration_unique_field

    def registration_raise_on_illegal_value_change(self, registered_subject):
        """Raises an exception if a value changes between
        updates.

        Values are available in `registration_options`.
        """
        pass

    @property
    def registration_options(self):
        """Gathers values for common attributes between the
        registration model and this instance.
        """
        registration_options = {}
        rs = self.registration_model()
        for k, v in self.__dict__.items():
            if k not in DEFAULT_BASE_FIELDS + ["_state"]:
                try:
                    getattr(rs, k)
                    registration_options.update({k: v})
                except AttributeError:
                    pass
        registration_identifier = registration_options.get("registration_identifier")
        if registration_identifier:
            registration_options["registration_identifier"] = self.to_string(
                registration_identifier
            )
        return registration_options

    class Meta:
        abstract = True



# Generated by Django 2.2.3 on 2019-09-17 18:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_registration", "0017_auto_20190305_0123")]

    operations = [
        migrations.AddField(
            model_name="historicalregisteredsubject",
            name="randomization_list_model",
            field=models.CharField(max_length=150, null=True),
        ),
        migrations.AddField(
            model_name="registeredsubject",
            name="randomization_list_model",
            field=models.CharField(max_length=150, null=True),
        ),
    ]
